# Todo: Define merge as '|' and cat as '+'
#
#
# time speed of using iadd operator.

# Music data representation in Python - Jim Bumgardner
# - loosely based on phrases in KeyKit by Tim Thompson
import re

justRatio = [1,8.5/8.0,9/8.0,9.5/8.0,5/4.0,4/3.0,45/32.0,
             3/2.0,25/16.0,27/16.0,12/9.0,15/8.0];

nnumbers = {'c':0, 'd':2,'e':4,'f':5,'g':7,'a':9,'b':11}

def mergeCmp(a,b):
  if (a.time == b.time):
    if (a.length == b.length):
      return a.chan - b.chan
    return a.length - b.length      
  return a.time - b.time


class note:
  def __init__(self,noteNbr=60,vol=63,time=0,length=96,chan=1):
    self.noteNbr = noteNbr
    self.vol = vol
    self.time = time
    self.length = length
    self.chan = chan

  def midiToFreq(self):
    return 55*2**((self.noteNbr-33)/12.0)


  def midiToFreqJ(self):
    nearestOctave = 12 * int(self.noteNbr/12)  # 60 for middle c
    nstep = self.noteNbr-nearestOctave
    return (55*2**((nearestOctave-33)/12.0)) * justRatio[nstep]

class phrase:
    phr = '';
    
    def notes(self):
        for n in self.phr:
          yield n

    def length(self):
      if (len(self.phr) > 0):
        return self.phr[-1].time + self.phr[-1].length
      else:
        return 0

    def __or__(self,p2):
      p1 = phrase(self.phr)
      p1.phr += p2.phr;
      p1.phr.sort(mergeCmp);
      return p1

    def __ior__(self,p2):
      self.phr += p2.phr;
      self.phr.sort(mergeCmp);
      return self

    def __add__(self,p2):
      t = phrase()
      ln = self.length()
      for n in p2.notes():
        n.time += ln
      t.phr = self.phr + p2.phr
      return t

    def __iadd__(self,p2):
      ln = self.length()
      for n in p2.notes():
        n.time += ln
      self.phr += p2.phr
      return self

    def setChan(self,v):
      for n in self.notes():
        n.chan = v

    # !!! if phr is not none, clone it here
    def __init__(self,phr=None):
      if (phr is None):
        phr = []
      self.phr = phr
     
    def fromString( st='' ):
        this = phrase()
        if (st == ''):
          return

        stime = 0
        cvol = 63
        cdur = 96
        coct = 3
        cchan = 1

        noteprefix = re.compile(r'([a-gpr])(\d+)?')
        options = re.compile(r'[ovdc\-\+]\d+')

        for chord in st.split(','):
            cary = chord.split(' ')
            maxd = -1
            for j in cary:
                m = noteprefix.match(j)
                if (m):
                    if (m.group(1)[0] == 'r'):
                        noteNbr = -1
                        if (m.group(2)):
                            cdur = int(m.group(2))
                    elif (m.group(1)[0] == 'p'):
                        noteNbr = int(m.group(1))
                    else:
                        noteNbr = nnumbers[m.group(1)]
                        if (m.group(2)):
                            coct = int(m.group(2))
                    j = j[m.end(0):]
                else:
                    print "Invalid Note: ",j
                    continue
                for opt in options.findall(j):
                    if (opt[0] == 'o'):
                        coct = int(opt[1:]);
                    elif (opt[0] == 'v'):
                        cvol = int(opt[1:]);
                    elif (opt[0] == 'd'):
                        cdur = int(opt[1:]);
                    elif (opt[0] == 'c'):
                        cchan = int(opt[1:]);
                    elif (opt[0] == '+'):
                        noteNbr += 1
                    elif (opt[0] == '-'):
                        noteNbr -= 1
                if (noteNbr != -1):
                  noteNbr += 12*(coct+2)
                this.phr.append(note(noteNbr,cvol,stime,cdur,cchan))
                if (maxd < cdur):
                    maxd = cdur
            stime += maxd
        return this

    fromString = staticmethod( fromString )