#!/usr/local/bin/perl -s

# Work on reducing blank answers...  1994_004 often generates a blank on 'who' questions

# one-hit-wonders
# who spent the most time at #1 on the charts in <year>?
# what year did # [first] hit #1 with _____ ?

# 
# Rank Difficulty of question by popularity of artist (based on # tracks) and answer track.
#
# Rank Difficulty of Answers (different artist = ez, same artist, close year = hard), so we can eliminate from easy->hard
#    different artist, genre, year...

#use Data::Dumper;
use strict;

$| = 1;

#$Data::Dumper::Terse = 1;  # avoids $VAR1 = * ; in dumper output
#$Data::Dumper::Indent = 0;  # more concise output


my @gameStyles = 
(
	{title=>"All Star Music Trivia", trackPicker=>\&GamePickerGeneral},
	{title=>"90s Music Trivia", trackPicker=>\&GamePicker90s},
	{title=>"80s Music Trivia", trackPicker=>\&GamePicker80s},
	{title=>"Oldies Music Trivia", trackPicker=>\&GamePickerOldies},
#	{title=>"Classic Rock Music Trivia", trackPicker=>\&GamePickerClassicRock},
	{title=>"R&B Music Trivia", trackPicker=>\&GamePickerRB},
	{title=>"Rap Music Trivia", trackPicker=>\&GamePickerRap},
	{title=>"Country Music Trivia", trackPicker=>\&GamePickerCountry},
	{title=>"Jazz/Vocal Music Trivia", trackPicker=>\&GamePickerJazz},
	{title=>"One-hit wonders", trackPicker=>\&GamePickerOneHit},
	{title=>"Famous Covers", trackPicker=>\&GamePickerCovers},
);

my $gameIndex = shift;
$gameIndex = int(rand() * @gameStyles) if $gameIndex eq '' || $gameIndex >= @gameStyles;

my $gameNumber = shift;
$gameNumber = 0 if !$gameNumber;

my $minRank = 40;  # use top 40s only... (use top10s for easy, top40s for medium)
my $cutOff = 10;   # must have at least this many hits...
my $nbrQuestions = 10;
my $debug = 0;

my $gameStyle = $gameStyles[$gameIndex];

my @qTypes = ();

if ($gameIndex != 9)
{
	@qTypes = (
		{qtemplate=>"Which <artist> hit stayed on the charts for <chartweeks> weeks in <year>?",
		 atemplate=>"<track>",
		 candFilter=>\&GoodSubject,
		 hardAltFilter=>\&GoodHardAlternate, hardPost=>\&SortByHigh,
		 easyAltFilter=>\&GoodEasyAlternate, easyPost=>\&SortByHigh
		},
		{qtemplate=>"This #<charthigh> hit by <artist> stayed on the charts for <chartweeks> weeks in <year>. What is it?",
		 atemplate=>"<track>",
		 candFilter=>\&GoodSubject,
		 hardAltFilter=>\&GoodHardAlternate, hardPost=>\&SortByHigh,
		 easyAltFilter=>\&GoodEasyAlternate, easyPost=>\&SortByHigh,
		},
		{qtemplate=>"Who hit #<charthigh> in <year> with <track>?",
		 atemplate=>"<artist>",
		 candFilter=>\&GoodSubjectNumber1,
		 hardAltFilter=>\&SameYearDifferentArtistSameGenreHit1,	 hardPost=>\&SortByHigh,
		 easyAltFilter=>\&CloseYearDifferentArtistHit2,	 easyPost=>\&SortByHighAndClose,
		 ansPost => \&UniqueArtists
		},
		{qtemplate=>"What year did <artist> hit #<charthigh> with <track>?",
		 atemplate=>"<year>",
		 candFilter=>\&GoodSubjectNumber1,
		 hardAltFilter=>\&DifferentYearSameArtist,	 	hardPost=>\&SortByHigh,
		 easyAltFilter=>\&CloseButDifferentYear,	 	easyPost=>\&UniqueYears,
		 ansPost => \&UniqueYears
		}
	);
}
else {
	@qTypes = (
		{qtemplate=>"Which <artist> cover (previously a hit for <originalartist>) stayed on the charts for <chartweeks> weeks in <year>?",
		 atemplate=>"<track>",
		 candFilter=>\&GoodSubject,
		 hardAltFilter=>\&GoodHardAlternate, hardPost=>\&SortByHigh,
		 easyAltFilter=>\&GoodEasyAlternate, easyPost=>\&SortByHigh
		},
		{qtemplate=>"<artist> had a #<charthigh> hit with this <originalartist> cover.  It stayed on the charts for <chartweeks> weeks in <year>. What is it?",
		 atemplate=>"<track>",
		 candFilter=>\&GoodSubject,
		 hardAltFilter=>\&GoodHardAlternate, hardPost=>\&SortByHigh,
		 easyAltFilter=>\&GoodEasyAlternate, easyPost=>\&SortByHigh,
		},
		{qtemplate=>"Who hit #<charthigh> in <year> with <track>, following <originalartist>?",
		 atemplate=>"<artist>",
		 candFilter=>\&GoodSubjectNumber1,
		 hardAltFilter=>\&SameYearDifferentArtistSameGenreHit1,	 hardPost=>\&SortByHigh,
		 easyAltFilter=>\&CloseYearDifferentArtistHit2,	 easyPost=>\&SortByHighAndClose,
		 ansPost => \&UniqueArtists
		},
		{qtemplate=>"What year did <artist> hit #<charthigh> with <track>, previously a hit for <originalartist>?",
		 atemplate=>"<year>",
		 candFilter=>\&GoodSubjectNumber1,
		 hardAltFilter=>\&DifferentYearSameArtist,	 	hardPost=>\&SortByHigh,
		 easyAltFilter=>\&CloseButDifferentYear,	 	easyPost=>\&UniqueYears,
		 ansPost => \&UniqueYears
		}
	);
}



my $idxYear = 0;
my $idxYearlyRank = 1; # ?? not sure what this signifies
my $idxPrefix = 2;     # this makes a unique-id
my $idxCH = 3;
my $idxHigh = 7;
my $idxArtist = 8;
my $idxFeatured = 9;
my $idxTrack = 10;
my $idxGenre = 11;
my $idxCover = 12;



#my $startTime = time();
require "BillboardData.pl";
my @data = getData();

#my %genreCnts = ();
#foreach my $rec (@data) {
#	$genreCnts{$rec->[$idxGenre]}++;
#}
#foreach my $key (sort {$genreCnts{$b} <=> $genreCnts{$a}} keys %genreCnts)
#{
#	printf "%5d %s\n", $genreCnts{$key}, $key;
#}
#exit;



splice @data,0,1;  # remove labels
# print Dumper($data[0]);



# Setup Data
#
my %artists = ();
for (my $y = 0; $y < $#data; ++$y)
{
	my $rec = $data[$y];
	next if $rec->[$idxYear] < 1955 || $rec->[$idxFeatured] ne '' || $rec->[$idxHigh] > $minRank;
	my $aName = $rec->[$idxArtist];
	$artists{$aName} = {tracks=>{}, nbrHits=>0} if not defined $artists{$aName};
	$artists{$aName}->{tracks}->{$rec->[$idxTrack]} = $y;
	$artists{$aName}->{nbrHits}++ if $rec->[$idxHigh] >= 1 && $rec->[$idxHigh] <= 40;
}

foreach my $aName (keys %artists)
{
	$artists{$aName}->{nbrTracks} = scalar(keys %{$artists{$aName}->{tracks}});
}

my %prevAnswers = ();
my @gameAnswers = ();

printf '<game title="%s" type="%d" number="%d">' . "\n", encodeStr($gameStyle->{title}), $gameIndex, $gameNumber;
printf "<topscores><score s=\"100\" nom=\"jbum\" ts=\"%d\"/></topscores>\n", time();

my $qNo = 1;
while ($qNo <= $nbrQuestions)
{
	my $q = makeQuestion($qTypes[$qNo % scalar(@qTypes)], $qNo); # pass in question variety data here...
	if ($q ne '') {
		printf "%s", $q->{txt};
		++$qNo;
	}
}
# printf "<answers>%s</answers>" , (join ',', @gameAnswers);
print "</game>\n";
exit 0;

sub makeQuestion($$)
{
	my ($qRec, $qNo) = @_;
	# Filter candidate tracks
	#
	my @candidates = grep {&{$qRec->{candFilter}}($_)} @data;

	my $tIdx = int(rand() * scalar(@candidates));
	my $track = $candidates[$tIdx];
	my $aName = $track->[$idxArtist];
	my $artist = $artists{$aName};

	$prevAnswers{$track->[$idxPrefix]} = 1;

	# get artist track names
	my @tKeys = keys %{$artist->{tracks}};
		
	# Make One Hard Alternate
	#
	my @alts = grep {$_->[$idxPrefix] ne $track->[$idxPrefix] && 
	                 &{$qRec->{hardAltFilter}}($_,$track)} @data;

	# post process hard alts
	if (scalar(@alts) && defined $qRec->{hardPost})
	{
		@alts = &{$qRec->{hardPost}}(\@alts);
	}
	@alts = splice @alts, 0, 1;

	# Make Two Easy Alternates
	#
	my @easyAlts = grep {$_->[$idxPrefix] ne $track->[$idxPrefix] && 
	                     $_->[$idxPrefix] != $alts[0]->[$idxPrefix] && 
		    			 &{$qRec->{easyAltFilter}}($_, $track)} @data;

	# post process easy alts
	if (scalar(@easyAlts) && defined $qRec->{easyPost})
	{
		@easyAlts = &{$qRec->{easyPost}}(\@easyAlts);
	}

	# add them to hard alts, post-process again if need be
	push @alts, @easyAlts;
	if (defined $qRec->{ansPost})
	{
		@alts = &{$qRec->{ansPost}}(\@alts);
	}
	@alts = grep { getAnswer($qRec, $_) ne ''} @alts;
	# printf "PROBLEM: Not enough ALTs\n" if (@alts < 3);
	return '' if (@alts < 3);

	my @answers = ($track, $alts[0], $alts[1], $alts[2]);
	my @ord = (0,1,2,3);
	
	my $q = $qRec->{qtemplate};
	$q =~ s/<artist>/Fixname($aName)/ge;
	$q =~ s/<year>/$track->[$idxYear]/ge;
	$q =~ s/<chartweeks>/$track->[$idxCH]/ge;
	$q =~ s/<charthigh>/$track->[$idxHigh]/ge;
	$q =~ s/<track>/$track->[$idxTrack]/ge;
	if ($track->[$idxCover] > 0) {
		my $prevTrack = $track;
		while ($prevTrack->[$idxCover] != 0)
		{
			$prevTrack = $data[$prevTrack->[$idxCover]-1];
		}
		$q =~ s/<originalartist>/Fixname($prevTrack->[$idxArtist])/ge;
	}
	my $genrespace = ($track->[$idxGenre] ne ''? $track->[$idxGenre].' ' : '');
	$q =~ s/<genrespace>/$genrespace/ge;
	my $txt = sprintf '<question q="%s" aname="%s" tname="%s">' . "\n", encodeStr($q), encodeStr(Fixname($aName)), encodeStr($track->[$idxTrack]);
	
	# Arrange Answers in Random Order
	for my $an (0..3) {
		my $rPick = int(rand() * scalar(@answers));
		my $atrack = splice @answers, $rPick, 1;
		my $aord = splice @ord, $rPick, 1;
		my $ansT = getAnswer($qRec, $atrack);
		$ansT =~ s/\"/%22/;
		$txt .= sprintf '<ans a="%s" o="%d" ', encodeStr($ansT), $aord;
		if ($atrack->[$idxPrefix] eq $track->[$idxPrefix]) 
		{
			$txt .= 'c="1" ';
			$gameAnswers[$qNo] = sprintf '%c', 65+$an;
		}
		$txt .= "/>\n";
	}
	$txt .= "</question>\n";
	return {txt=>$txt};
}


sub encodeStr($)
{
	my ($str) = @_;
	$str =~ s/\"/%22/g;
	$str =~ s/\&/&amp;/g;
	return $str;
}

sub getAnswer($$)
{
	my ($qRec, $atrack) = @_;
	my $aa = $qRec->{atemplate};
	$aa =~ s/<artist>/Fixname($atrack->[$idxArtist])/ge;
	$aa =~ s/<track>/$atrack->[$idxTrack]/ge;
	$aa =~ s/<year>/$atrack->[$idxYear]/ge;
	$aa =~ s/<charthigh>/$atrack->[$idxHigh]/ge;
	$aa =~ s/<chartweeks>/$atrack->[$idxCH]/ge;
	return $aa;
}

sub GamePickerGeneral($)
{
	my ($track) = @_;
	my $year = $track->[$idxYear];
	return 0 if $artists{$track->[$idxArtist]}->{nbrTracks} < 10;
	if (rand() < .5) {
		return $year > 1998;
	}
	elsif (rand() < .5) {
		return $year > 1988;
	}
	elsif (rand() < .5) {
		return $year > 1978;
	}
	elsif (rand() < .5) {
		return $year > 1968;
	}
	return $year > 1958;
}

sub GamePicker90s($)
{
	my ($track) = @_;
	my $year = $track->[$idxYear];
	return 0 if $artists{$track->[$idxArtist]}->{nbrTracks} < 5;
	return $year >= 1990 && $year <= 1999;
}

sub GamePicker80s($)
{
	my ($track) = @_;
	my $year = $track->[$idxYear];
	return 0 if $artists{$track->[$idxArtist]}->{nbrTracks} < 5;
	return $year >= 1980 && $year <= 1989;
}

sub GamePickerOldies($)
{
	my ($track) = @_;
	my $year = $track->[$idxYear];
	return 0 if $artists{$track->[$idxArtist]}->{nbrTracks} < 5;
	return $year >= 1956 && $year <= 1979;
}

sub GamePickerClassicRock($)
{
	my ($track) = @_;
	my $year = $track->[$idxYear];
	return 0 if $artists{$track->[$idxArtist]}->{nbrTracks} < 5;
	return $year >= 1967 && $year <= 1979 && $track->[$idxGenre] eq "Rock";
}

sub GamePickerRB($)
{
	my ($track) = @_;
	my $year = $track->[$idxYear];
	return 0 if $track->[$idxGenre] ne 'R&B';
	if (rand() < .5) {
		return $year > 1998;
	}
	elsif (rand() < .5) {
		return $year > 1988;
	}
	elsif (rand() < .5) {
		return $year > 1978;
	}
	elsif (rand() < .5) {
		return $year > 1968;
	}
	return 1;
}

sub GamePickerRap($)
{
	my ($track) = @_;
	my $year = $track->[$idxYear];
	return 0 if $track->[$idxGenre] ne 'Rap';
	return 1 if $year > 1970;
	return 0;
}

sub GamePickerCountry($)
{
	my ($track) = @_;
	my $year = $track->[$idxYear];
	return 0 if $track->[$idxGenre] ne 'Country';
	return 1 if $year > 1956;
	return 0;
}

sub GamePickerJazz($)
{
	my ($track) = @_;
	my $year = $track->[$idxYear];
	return 0 if $track->[$idxGenre] ne 'Jazz' && $track->[$idxGenre] ne 'Vocal';
	return 1 if $year > 1930;
	return 0;
}

sub GamePickerOneHit($)
{
	my ($track) = @_;
	my $year = $track->[$idxYear];
	return 0 if $year < 1956 || $year > 2003;
    return 0 if $artists{$track->[$idxArtist]}->{nbrHits} > 1;
    return 0 if $track->[$idxHigh] > 20;
	return 1;
}

sub GamePickerCovers($)
{
	my ($track) = @_;
	my $year = $track->[$idxYear];
	return 0 if $year < 1958;
	return 0 if $track->[$idxCover] == 0;
    return 0 if $track->[$idxHigh] > 20;
	return 1;
}

# print @$_ foreach @data;
sub GoodSubject($)
{
	my ($track) = @_;
	return 1 if &{$gameStyle->{trackPicker}}($track) && 
	             (not defined $prevAnswers{$track->[$idxPrefix]} ) &&
	             $track->[$idxHigh] >= 1 &&
	             $track->[$idxHigh] <= 2 &&
	             $track->[$idxFeatured] eq '';
	return 0;
}

sub GoodSubjectNumber1($)
{
	my ($track) = @_;
	return 1 if &{$gameStyle->{trackPicker}} && 
	            (not defined $prevAnswers{$track->[$idxPrefix]} ) &&
	            $track->[$idxHigh] == 1 &&
	            $track->[$idxFeatured] eq '';
	return 0;
}

sub GoodHardAlternate($$)
{
	my ($track, $qTrack) = @_;
	return 1 if $track->[$idxArtist] eq $qTrack->[$idxArtist] &&
	            $track->[$idxYear] != $qTrack->[$idxYear];
	return 0;
	# same artist, different year, sort by popularity
}

sub SameYearDifferentArtistSameGenreHit1($$)
{
	my ($track, $qTrack) = @_;
	return 1 if $track->[$idxYear] == $qTrack->[$idxYear] &&
				$track->[$idxArtist] ne $qTrack->[$idxArtist] &&
	            ($track->[$idxGenre] eq $qTrack->[$idxGenre] || $track->[$idxGenre] eq '' || $qTrack->[$idxGenre] eq '') &&
	            $track->[$idxHigh] <= 3;
	return 0;
}

sub CloseYearDifferentArtistHit2($$)
{
	my ($track, $qTrack) = @_;
	return 1 if abs($track->[$idxYear] - $qTrack->[$idxYear]) < 5 &&
				$track->[$idxArtist] ne $qTrack->[$idxArtist] &&
	            ($track->[$idxGenre] eq $qTrack->[$idxGenre] || $track->[$idxGenre] eq '' || $qTrack->[$idxGenre] eq '') &&
	            $track->[$idxHigh] <= 3;
	return 0;
}


# should always produce complimentary set to hard, to avoid repeats
sub GoodEasyAlternate($$)
{
	my ($track, $qTrack) = @_;
	return 1 if $track->[$idxYear] == $qTrack->[$idxYear] &&
				$track->[$idxArtist] ne $qTrack->[$idxArtist] &&
	            $track->[$idxCH] >= 10 &&
	            ($track->[$idxGenre] eq $qTrack->[$idxGenre] || $track->[$idxGenre] eq '' || $qTrack->[$idxGenre] eq '');
	return 0;
}

sub DifferentYearSameArtist($$)
{
	my ($track, $qTrack) = @_;
	return 1 if abs($track->[$idxYear] != $qTrack->[$idxYear]) &&
				$track->[$idxArtist] eq $qTrack->[$idxArtist];
	return 0;
}

sub CloseButDifferentYear($$)
{
	my ($track, $qTrack) = @_;
	return 1 if abs($track->[$idxYear] != $qTrack->[$idxYear]) &&
	            abs($track->[$idxYear] - $qTrack->[$idxYear]) < 3;
	return 0;
}

sub SortByHigh($)
{
	my ($ary) = @_;
	my @nary = sort {$a->[$idxHigh] <=> $b->[$idxHigh]} @{$ary};
	return @nary;
}

sub SortByHighAndClose($)
{
	#!!! also sort close in distance to top... need to pass target track
	my ($ary) = @_;
	my @nary = sort {$a->[$idxHigh] <=> $b->[$idxHigh]} @{$ary};
	return @nary;
}

sub UniqueArtists($)
{
	my ($ary) = @_;
	my %uniqArtists = ();
	my @nAry = ();
	foreach my $trk (@{$ary})
	{
		if (not defined $uniqArtists{$trk->[$idxArtist]})
		{
			$uniqArtists{$trk->[$idxArtist]} = 1;
			push @nAry, $trk;
		}
	}
	return @nAry;
}

sub UniqueYears($)
{
	my ($ary) = @_;
	my %uniqYears = ();
	my @nAry = ();
	foreach my $trk (@{$ary})
	{
		if (not defined $uniqYears{$trk->[$idxYear]})
		{
			$uniqYears{$trk->[$idxYear]} = 1;
			push @nAry, $trk;
		}
	}
	return @nAry;
}

sub Fixname($)
{
	my ($nom) = @_;
	$nom = "The " . $1 if $nom =~ /(.*), The$/i;
	return $nom;
}