#!/usr/local/bin/perl -s

# Work on reducing blank answers...  1994_004 often generates a blank on 'who' questions

# one-hit-wonders
# who spent the most time at #1 on the charts in <year>?
# what year did # [first] hit #1 with _____ ?

# 
# Rank Difficulty of question by popularity of artist (based on # tracks) and answer track.
#
# Rank Difficulty of Answers (different artist = ez, same artist, close year = hard), so we can eliminate from easy->hard
#    different artist, genre, year...

use Text::CSV::Simple;
use Data::Dumper;
use strict;

$| = 1;

$Data::Dumper::Terse = 1;  # avoids $VAR1 = * ; in dumper output
$Data::Dumper::Indent = 0;  # more concise output

my $minRank = 40;  # use top 40s only... (use top10s for easy, top40s for medium)
my $cutOff = 10;   # must have at least this many hits...
my $nbrQuestions = 10;
my $debug = 0;

my $startTime = time();
my $infile = shift;
my $infile = "BillboardPopME.csv" if !$infile;
my $parser = Text::CSV::Simple->new;
$parser->want_fields(0,1,2,3,4,5,6,7,9,11,12,17,19);
my $datafile = $infile;
my @data = $parser->read_file($datafile);
printf "%d seconds elapsed\n", time()- $startTime;

my $idxYear = 0;
my $idxYearlyRank = 1; # ?? not sure what this signifies
my $idxPrefix = 2;     # this makes a unique-id
my $idxCH = 3;
my $idxHigh = 7;
my $idxArtist = 8;
my $idxFeatured = 9;
my $idxTrack = 10;
my $idxGenre = 11;
my $idxWrittenBy = 12;
my $idxCover = 12;

print "Finding covers...\n";
for (my $y = 0; $y < $#data; ++$y)
{
	my $rec = $data[$y];
	print "\r$y" if $y % 1000 == 0;
	my $writtenBy = $rec->[$idxWrittenBy];
	$rec->[$idxWrittenBy] = 0;
	next if $writtenBy eq '' || $rec->[$idxHigh] > 20;
	for (my $z = $y+1; $z < $#data; ++$z)
	{
		my $rec2 = $data[$z];
		last if $rec2->[$idxYear] < 1958;
		if (lc($rec->[$idxTrack]) eq lc($rec2->[$idxTrack]) &&
		    lc($writtenBy) eq lc($rec2->[$idxWrittenBy]) &&
		    lc($rec->[$idxArtist]) ne lc($rec2->[$idxArtist]))
		{
			$rec->[$idxCover] = $z;
		}
	}
}	
print "\n";


open (OFILE, ">BillboardData.pl");
print OFILE <<EOT;
sub getData()
{
my \@myData = (
EOT
foreach my $rec (@data) {
	print OFILE Dumper($rec) . ",\n";
}
print OFILE <<EOT;
);

return \@myData;
}

1;
EOT
close OFILE;
exit;
